from PyQt5.QtCore import QEvent, QPoint, QTimer, Qt
from PyQt5.QtGui import QIcon, QImage, QPainter
from PyQt5.QtWidgets import QApplication, QHBoxLayout, QMainWindow, QTabWidget, QWidget
from lib import constants
from lib.constants import ResPath
from lib.i18n import lu, luDir
from util import utils
from util.simulationcontext import SimulationContext
from widgets.creditstab import CreditsLabel
from widgets.documentation import DocumentationTabWidget
from widgets.materialeditor.materialeditortab import MaterialEditorWidget
from widgets.overlaywidget import OverlayWidget
from widgets.simulation.inversiontab import InversionWidget
from widgets.simulation.simulationtab import AnimationWidget
from widgets.testwidget import TestWidget
import ctypes
import platform
import sys


class Window(QMainWindow):
    labelClickHint = lu("labelClickHint")

    def __init__(self):
        QMainWindow.__init__(self)
        self.introductionImage = QImage(luDir(ResPath.documentation) + "overview_overlay.png")
        self.speechBubbleImage = QImage(ResPath.textures + "speechBubble.png")

        self.setWindowTitle(lu("title"))
        self.setWindowIcon(QIcon(constants.ResPath.icons + "logo.png"))
        # prevent the firing of a resize event (that gets called by the layout later)
        self.blockSignals(True)
        self.setFixedSize(constants.DEFAULT_WIDTH, constants.DEFAULT_HEIGHT)
        self.blockSignals(False)

        self._simulationContext = SimulationContext()

        dummy = QWidget()
        mainLayout = QHBoxLayout(dummy)
        self.setCentralWidget(dummy)

        #######################################################
        # on left side is the tab widget for all interactions #
        #######################################################
        self._interactionTab = QTabWidget()
        mainLayout.addWidget(self._interactionTab, stretch=2)
        # material editor widget
        self._materialEditor = MaterialEditorWidget(self)
        self._interactionTab.addTab(self._materialEditor, lu("tabMaterialProperties"))
        # animation widget
        self._animWidget = AnimationWidget(self)
        self._interactionTab.addTab(self._animWidget, lu("tabAnimation"))
        self._interactionTab.setTabEnabled(1, False)
        # inversion tab
        self._inversionWidget = InversionWidget(self)
        self._interactionTab.addTab(self._inversionWidget, lu("tabInversion"))
        self._interactionTab.setTabEnabled(2, False)
        # credits tab
        self._interactionTab.addTab(CreditsLabel(), lu("tabCredits"))
        # test widget
        #self._interactionTab.addTab(TestWidget(self), "Test")

        #########################################################################
        # on the right side is the tab widget for all documentation (tut, math) #
        #########################################################################
        self._documentationTab = DocumentationTabWidget(self)
        mainLayout.addWidget(self._documentationTab, stretch=1)
       

        ###########
        # Overlay #
        ###########
        self._overlay = OverlayWidget(dummy)
        self._overlay.setHidden(True)

        self.forceTabUpdate(self._interactionTab)
        
        self.keptInitialMaterial = None
        


    def postInit(self):
        bWidth = 256
        self.speechBubbleImage = self.speechBubbleImage.scaledToWidth(bWidth, mode=Qt.SmoothTransformation)
        # subtract 10 from width because introImage will be placed at x=10
        # subtract 'bWidth+32' from width so that clicking hint will never overlap
        self.introductionImage = self.introductionImage.scaled(self.width() - 10 - (bWidth + 32), self.height(),
                                                               aspectRatioMode=Qt.KeepAspectRatio, transformMode=Qt.SmoothTransformation)
        self.showIntroductionOverlay()

    def forceTabUpdate(self, tabWidget):
        self.setAttribute(Qt.WA_DontShowOnScreen)
        self.show()

        for i in range(1, tabWidget.count()):
            tabWidget.setCurrentIndex(i)

        self.hide()
        self.setAttribute(Qt.WA_DontShowOnScreen, on=False)
        tabWidget.setCurrentIndex(0)

    @property
    def simulationContext(self):
        return self._simulationContext

    def switchToSimulation(self):
        fwdResult = self._simulationContext.fwdResult
        if not fwdResult:
            raise AttributeError()

        self._interactionTab.setTabEnabled(2, False)

        self._animWidget.reset()
        self._animWidget.loadFromCurrentContext()
        self._animWidget.start()
    
        # switch tab
        self._interactionTab.setTabEnabled(1, True)
        self._interactionTab.setCurrentIndex(1)
        # switch documentation page TODO maybe don't hard-code this
        self._documentationTab.setCurrentIndex(0)
        self._documentationTab._docWidgetTut.goTo(2)

#     def switchToSimulation(self, FWD, signalPlacementImage, material, background_material, shot, measurements,
#                            source_position, receivers_positions,
#                            sceneImage, frame_generator, seismogram_image):
#         # TODO: wohin damit?
#         self._FWD = FWD
#         self._shot = shot
#         self._material = material
#         self._initial_material = background_material
#         self._measurements = measurements
#         self._source_position = source_position
#         self._receivers_positions = receivers_positions
#         self._signalPlacementImage = signalPlacementImage
#         self._animWidget.reset()
#         # set the background (scene) Image of the animation widget
#         self._animWidget.setBackground(sceneImage)
#         # set the animation frames
#         self._animWidget.setFrameGenerator(frame_generator)
#         self._animWidget.setSeismogram(seismogram_image)
#         self._animWidget.start()
#
#         # switch tab
#         self._interactionTab.setTabEnabled(1, True)
#         self._interactionTab.setCurrentIndex(1)

    def switchToInversion(self):
        self._animWidget.pause()
        
        self._inversionWidget.reset()
        self._inversionWidget.loadFromCurrentContext(self.keptInitialMaterial)
        self.keptInitialMaterial = None
        # switch tab
        self._interactionTab.setTabEnabled(2, True)
        self._interactionTab.setCurrentIndex(2)
        # switch documentation page TODO maybe don't hard-code this
        self._documentationTab.setCurrentIndex(0)
        self._documentationTab._docWidgetTut.goTo(3)

#     def switchToInversion(self, sceneImage):  # , realMaterial, fwd, measurements):
#         self._inversionWidget.reset()
#         self._inversionWidget.initialize(sceneImage, self._signalPlacementImage, self._FWD, self._shot,
#                                          self._source_position, self._receivers_positions,
#                                          self._material, self._material, self._measurements)
#                                         # self._material, self._initial_material, self._measurements)
#         # switch tab
#         self._interactionTab.setTabEnabled(2, True)
#         self._interactionTab.setCurrentIndex(2)

    def switchToMaterialEditor(self):
        self._interactionTab.setCurrentIndex(0)
        self._documentationTab._docWidgetTut.goTo(1)

    def getOverlay(self):
        return self._overlay

    def resizeEvent(self, event):
        QMainWindow.resizeEvent(self, event)
        self._overlay.resize(event.size())
        event.accept()

    def reset(self):
        print "reset"
        self._interactionTab.setTabEnabled(1, False)
        self._interactionTab.setTabEnabled(2, False)
        self.switchToMaterialEditor()
        
        for i in range(self._interactionTab.count()):
            tabWidget = self._interactionTab.widget(i)
            tabWidget.reset()

        self._documentationTab.setCurrentIndex(0)
        for i in range(self._documentationTab.count()):
            tabWidget = self._documentationTab.widget(i)
            tabWidget.home()

        self.showIntroductionOverlay()

    def showIntroductionOverlay(self):
        self._overlay.hide(animate=False)

        # add intro image
        x = 10 # (self.width() - image.width()) // 2
        self._overlay.addImages((x, 0), self.introductionImage)
        # add clicking hint
        x = self.width() - (self.speechBubbleImage.width() + 32)
        self._overlay.addImages((x, 96), self.speechBubbleImage)
        text = Window.labelClickHint
        textWidth = self._overlay.getStringWidth(text, 18)
        # center text
        x += (self.speechBubbleImage.width() - textWidth) // 2
        y = 96 + (self.speechBubbleImage.height() + 18) // 2
        self._overlay.addText((x, y), text, textSize=18, color=(0, 0, 0))
        self._overlay.show(animate=False, opacity=240)

    def cleanup(self):
        self._inversionWidget.cleanup()


class Application(QApplication):
    _resetEvents = (QEvent.KeyPress, QEvent.KeyRelease, QEvent.MouseButtonPress, QEvent.MouseButtonRelease, QEvent.MouseMove, QEvent.Wheel)


    def __init__(self, *args, **kwargs):
        QApplication.__init__(self, *args, **kwargs)

        self._idleWarningText = lu("idleWarning")

        self._idleWarningTimer = None
        self._idleOverlayTimer = None
        self._ignoreFlag = False

    def postInit(self, window):
        self._window = window
        window.postInit()

        self._idleWarningTimer = QTimer(self, timeout=self.resetWarning)
        self._idleWarningTimer.setSingleShot(True)
        self._idleWarningTimer.setInterval(constants.MAX_IDLE_TIME)

        self._idleOverlayTimer = QTimer(self, timeout=self.reset)
        self._idleOverlayTimer.setSingleShot(True)
        self._idleOverlayTimer.setInterval(constants.IDLE_WARNING_TIME)

        self._idleWarningTimer.start()

    def notify(self, receiver, event):
        if not self._ignoreFlag and event.type() in Application._resetEvents:
            self.resetTimer()
        return QApplication.notify(self, receiver, event)

    def resetTimer(self):
        if self._idleWarningTimer:
            if self._idleOverlayTimer and self._idleOverlayTimer.isActive():
                self._idleOverlayTimer.stop()
                self._window.getOverlay().hide()
            self._idleWarningTimer.start()

    def resetWarning(self):
        x, y = self._window.width() / 3, self._window.height() / 3
        self._ignoreFlag = True
        self._window.getOverlay().hide(animate=False)
        self._window.getOverlay().addText((QPoint(x, y),), (self._idleWarningText,))
        self._window.getOverlay().show()
        self.processEvents()
        self._ignoreFlag = False
        self._idleOverlayTimer.start()

    def reset(self):
        self._window.reset()
        # (re-) start idle timer
        self.resetTimer()



if __name__ == '__main__':
    # to show exceptions from Qt after crash
    sys.excepthook = utils.excepthook

    # create the QApplication
    app = Application(sys.argv)
    # tell windows that pythonw.exe is just a host for this app, so that the correct icon is shown in the task bar
    if platform.system() == "Windows":
        appId = u'kit_ianm.fwi_competition'  # arbitrary string
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(appId)

    # create the main window
    window = Window()
    app.postInit(window)
    window.show()
    exitCode = app.exec_()
    window.cleanup()
    sys.exit(exitCode)
