import copy
import math
from multiprocessing import Process, Queue, Lock
import threading

from lib.constants import ResPath, DEBUG
import numpy as np
import util.renderutil as renderutil
from util.simulationcontext import InversionResult


class InverseSolverThread(threading.Thread):
    def __init__(self, context, initial_material, *args, **kwargs):        
        threading.Thread.__init__(self)
        self.inv_solve = InverseSolver(context, initial_material, *args, **kwargs)
        # self.daemon = True
        self.run = self.inv_solve.run
        self.stop = self.inv_solve.stop
        
    def finished(self):
        return self.inv_solve.finished()
        
        # (ctx.fwdResult.shot, initial_material, ctx.FWDSolver, ctx.emitterPosition, ctx.receiverPositions)
        
class InverseSolver():
    def __init__(self, context, initial_material, max_iterations=4, wavelet_type="Ricker", source_fct=None):        
        # self.context = context 
        self._width = context.gridWidth
        self._waterHeight = context.waterHeight
        self._simulationHeight = context.simulationHeight
        
        self._shot = context.fwdResult.shot
        self._fwd = context.FWDSolver
        self._source_position = context.emitterPosition
        self._receiver_positions = context.receiverPositions
        # TODO take context source fct?
        self._wavelet_type = wavelet_type
        self._source_fct = source_fct
        
        self._max_iterations = max_iterations
        self._current_m = initial_material.copy()
        self._current_measurements = None
        self._current_gradient = None
        self._current_solution = None
        self._k = 0
        self._running = False
        self._finished = False
        
        self.inv_result = InversionResult()
                        
        self._lock = Lock()
                
        
    def material2image(self, m):
        (material_frame,) = self._fwd.animate([m], frame_count=1, colormap="binary")
        return material_frame
    
    def animateWave(self, wavefields, q):
        frames = self._fwd.animate(wavefields, frame_count=30)
        q.put(frames)
        
    def stop(self):
        with self._lock:
            self._running = False
    
    def run(self):
        with self._lock:
            self._running = True
        
         # store water material
        #water_material = self._current_m.reshape(self._width, self._simulationHeight)[:, :self._waterHeight]
                
        while self._running and self._k < self._max_iterations:
            if self._running:
                fwd_wavefields, measurements, shot, retval = \
                    self._fwd.solve(self._current_m,
                                    self._source_position, self._receiver_positions,
                                    self._wavelet_type, self._source_fct)
                fwd_solution = retval['dWaveOp']
                
                # update state
                with self._lock:
                    self._current_measurements, self._current_solution = \
                        measurements, fwd_solution
                
                self.inv_result.measurement_iterates[self._k] = copy.deepcopy(measurements)
                self.inv_result.forward_solutions[self._k] = copy.deepcopy(fwd_wavefields)
            
                # fwd_frames = self._fwd.animate(fwd_wavefields, frame_count=30)
                # for i,f in enumerate(fwd_frames):
                #    f.save("/tmp/fwd_{:03d}.png".format(i))
                    
                # self._fwd.plotSeismogram(measurements, (800,300)).save("/tmp/seismogram_{:03d}.png".format(self._k))
            
            if self._running: 
                gradient, bwd_wavefields = self._fwd.backpropAndGrad(self._shot,
                                                                     self._current_m, self._current_measurements,
                                                                     self._current_solution)
                    
                # don't change water material
                gradient.reshape(self._width, self._simulationHeight)[:, :self._waterHeight] = 0.0
                    
                # update state
                with self._lock:
                    self._current_gradient = gradient
                    
                self.inv_result.backward_solutions[self._k] = copy.deepcopy(bwd_wavefields)
                
                #self.material2image(gradient).save("/tmp/gradient_{:03d}.png".format(self._k))
                
                # bwd_frames = self._fwd.animate(bwd_wavefields, frame_count=30)
                # for i,f in enumerate(bwd_frames):
                #    f.save("/tmp/bwd_{:03d}.png".format(i))
            
            if self._running:
                alpha = self._fwd.caluclatealpha(self._shot, self._current_m, self._current_gradient,
                                                 self._current_solution)
        
                #print "alpha", alpha
                step = 10 * alpha*self._current_gradient
                
                # Warning: in current_m we store 1/c^2, but step contains a perturbation in c
                new_material = 1.0/(1.0/np.sqrt(self._current_m) + step)**2
                #print "stepnorm", np.sqrt(np.sum(step**2))
                #print "old materialnorm", np.sqrt(np.sum(self._current_m**2))
                #print "new materialnorm", np.sqrt(np.sum(new_material**2))
                
                # reset water
                #np.reshape(new_material, (self._width, self._simulationHeight))[:, :self._waterHeight] = water_material
                
                # update state
                with self._lock:
                    self._current_m = new_material
                    self.current_material = copy.deepcopy(self._current_m)

                # TODO better solution for subtracting base = 10
                new_mat_reshaped = new_material.reshape(self._width, self._simulationHeight)[:, self._waterHeight:] - 10 # 13 = waterheight, 10 = base                
                                              
                mat_img = renderutil.arrayToQImage(new_mat_reshaped)
                self.inv_result.material_images[self._k] = mat_img
                self.inv_result.material_iterates[self._k] = self._current_m.copy()
                
                #mat_img.save("/tmp/" + "mat_{:03d}.png".format(self._k))
                #mat_img_water = renderutil.arrayToQImage(new_material.reshape(self._width, self._simulationHeight) - 10) 
                #mat_img_water.save("/tmp/" + "mat_water{:03d}.png".format(self._k))
                
                if DEBUG:
                    mat_img.save(ResPath.tmp + "mat_{:03d}.png".format(self._k))
                    
                    np.set_printoptions(precision=1, threshold=20000, edgeitems=1000000, linewidth=14000)
                    with open(ResPath.tmp + "materialReconstruction_{:03d}.txt".format(self._k), "w") as f:
                        f.write(str(new_mat_reshaped))
                    mat_img_water = renderutil.arrayToQImage(new_material.reshape(self._width, self._simulationHeight) - 10) 
                    mat_img_water.save(ResPath.tmp + "mat_water{:03d}.png".format(self._k))
                                    
                
            with self._lock:
                self._k += 1
                print("finished iteration {}".format(self._k))
        
        with self._lock:
            self._running = False
            
        with self._lock:
            self._finished = True
        print "Inversion finished!"
        
    def finished(self):
        with self._lock:
            return self._finished
