from __future__ import absolute_import
import numpy as np
import matplotlib
from util.decorators import benchmark
from util.colormaps import CenteredColorMap
from PyQt5.QtGui import QImage
from util import renderutil
matplotlib.use("Agg")
import matplotlib.pyplot as plt
import matplotlib.figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas

from pysit.core.domain import RectangularDomain, PML
from pysit.core.mesh import CartesianMesh
from pysit.core.sources import PointSource
from pysit.core.wave_source import RickerWavelet, GaussianPulse, DerivativeGaussianPulse, SourceWaveletBase, _arrayify
from pysit.core.receivers import PointReceiver
from pysit.core.shot import Shot

from pysit.core.receivers import ReceiverSet
from pysit.solvers.wave_factory import ConstantDensityAcousticWave
from pysit.objective_functions.temporal_least_squares import TemporalLeastSquares

import pysit.vis as vis

from lib.constants import ResPath, DEBUG
from scipy.ndimage import filters

# constants
TOTAL_GRID_SIZE = 7000
WATER_MATERIAL = 10
BACKGROUND_MATERIAL = 13

def perturb_material(material, water_height=10):
    
    # TODO find better ways to perturb.
    
    # rand = np.random.standard_normal(size=material.shape)*2.0/15+1
    # material =  material*rand
    
    tmp_material = material.copy()
    #tmp_material[:, water_height:] = filters.gaussian_filter(material[:, water_height:], sigma=12, order=0)
    tmp_material[:, water_height:] = BACKGROUND_MATERIAL
    
    return tmp_material 
    # return material


def calculateGridSize(width, water_height, bottom_height, N=TOTAL_GRID_SIZE):
    total_height = water_height + bottom_height
    r = float(total_height) / width
    w = round(np.sqrt(N / r))
    h_total = round(r * w)
    h_water = round(water_height * h_total / total_height)
    h_bottom = round(bottom_height * h_total / total_height)
    return (int(w), int(h_water), int(h_bottom))


class CustomSource(SourceWaveletBase):
    def __init__(self, scaling, source_fct, *args, **kwargs):
        self._scaling = scaling
        self._source_fct = source_fct

    @property
    def time_source(self):
        """bool, Indicates if wavelet is defined in time domain."""
        return True

    def _evaluate_time(self, ts):
        # Vectorize the time list
        ts_was_not_array, ts = _arrayify(ts)

        # v = np.ones_like(ts)
        # v = np.sin(5*np.pi*ts)
        v = self._scaling * self._source_fct(ts / 10.0)
        # v[np.abs(v) < self.threshold] = 0.0

        # print(ts, v)

        return v[0] if ts_was_not_array else v

    def _evaluate_frequency(self, nus):
        raise NotImplementedError


class ForwardSolver:
    def __init__(self, dimensions=((-1, 1), (-1, 1)), grid_resolution=(30, 30), trange=(0, 1.7),
                 background_velocity=1.0):

        self.Nx, self.Ny = grid_resolution

        min_x, max_x = dimensions[0]
        min_y, max_y = dimensions[1]

        length_x = max_x - min_x
        length_y = max_y - min_y

        self._pmlx = PML(0.1 * length_x, 100)
        self._pmly = PML(0.1 * length_y, 100)

        x_config = (min_x, max_x, self._pmlx, self._pmlx)
        y_config = (min_y, max_y, self._pmly, self._pmly)
        # y_config = (min_y,max_y, Dirichlet(), self._pmly)

        self._background_velocity = background_velocity
        self._trange = trange

        self._domain = RectangularDomain(x_config, y_config)
        self._mesh = CartesianMesh(self._domain,
                                   self.Nx, self.Ny)

        self._initial_material = self._background_velocity * np.ones(self._mesh.shape())
        self._true_material = self._initial_material.copy()

        self._solver = ConstantDensityAcousticWave(self._mesh,
                                                   spatial_accuracy_order=8,
                                                   trange=self._trange,
                                                   kernel_implementation='cpp',
                                                   # cfl_safety=0.1,
                                                   formulation='scalar',
                                                   model_parameters={'C': self._true_material}
                                                   )
        self._objective = TemporalLeastSquares(self._solver) 

    def createShot(self, source_position, receiver_positions,
                   wavelet_type="Ricker", source_fct=None):
        wavelet_types = {"Ricker" : RickerWavelet(10.0),
                         "Gaussian" : GaussianPulse(10.0),
                         "dGaussian" : DerivativeGaussianPulse(10.0)}

        receivers = ReceiverSet(self._mesh,
                                [PointReceiver(self._mesh, p) for p in receiver_positions])

        if wavelet_type != "custom":
            wavelet = wavelet_types[wavelet_type]
        else:
            wavelet = CustomSource(10.0, source_fct)

        source = PointSource(self._mesh, source_position, wavelet)
        shot = Shot(source, receivers)
        return shot
    
    
    def _setup_forward_rhs(self, rhs_array, data):
        return self._solver.mesh.pad_array(data, out_array=rhs_array)
    
    def _forward_model(self, shot, m0, return_parameters=[], update_shot=False):
        """ orignal copied from modeling/temporal_modeling.py """

        # Local references
        solver = self._solver
        solver.model_parameters = m0
        
        mesh = solver.mesh
        
        d = solver.domain
        dt = solver.dt
        nsteps = solver.nsteps
        source = shot.sources

        # Storage for the field
        if 'wavefield' in return_parameters:
            us = list()
        
        # Setup data storage for the forward modeled data
        if 'simdata' in return_parameters:
            simdata = np.zeros((solver.nsteps, shot.receivers.receiver_count))            

        # Storage for the time derivatives of p
        if 'dWaveOp' in return_parameters:
            dWaveOp = list()
        
        # Step k = 0
        # p_0 is a zero array because if we assume the input signal is causal
        # and we assume that the initial system (i.e., p_(-2) and p_(-1)) is
        # uniformly zero, then the leapfrog scheme would compute that p_0 = 0 as
        # well. ukm1 is needed to compute the temporal derivative.
        solver_data = solver.SolverData()
        
        rhs_k   = np.zeros(mesh.shape(include_bc=True))
        rhs_kp1 = np.zeros(mesh.shape(include_bc=True))
        
        for k in xrange(nsteps):
        
            uk = solver_data.k.primary_wavefield
            uk_bulk = mesh.unpad_array(uk)
            
            if 'wavefield' in return_parameters:
                us.append(uk_bulk.copy())
                
            # Record the data at t_k
            if 'simdata' in return_parameters:
                shot.receivers.sample_data_from_array(uk_bulk, k, data=simdata)
                
            if update_shot:
                shot.receivers.sample_data_from_array(uk_bulk, k)
            
            if k == 0:
                rhs_k = self._setup_forward_rhs(rhs_k, source.f(k*dt))
                rhs_kp1 = self._setup_forward_rhs(rhs_kp1, source.f((k+1)*dt))
            else:
                # shift time forward
                rhs_k, rhs_kp1 = rhs_kp1, rhs_k
            rhs_kp1 = self._setup_forward_rhs(rhs_kp1, source.f((k+1)*dt))
            
            # Note, we compute result for k+1 even when k == nsteps-1.  We need
            # it for the time derivative at k=nsteps-1.
            solver.time_step(solver_data, rhs_k, rhs_kp1)

            # Compute time derivative of p at time k
            # Note that this is is returned as a PADDED array
            if 'dWaveOp' in return_parameters:
                dWaveOp.append(solver.compute_dWaveOp('time', solver_data))
        
            # When k is the nth step, the next step is uneeded, so don't swap 
            # any values.  This way, uk at the end is always the final step
            if(k == (nsteps-1)): break
            
            # Don't know what data is needed for the solver, so the solver data
            # handles advancing everything forward by one time step.
            # k-1 <-- k, k <-- k+1, etc
            solver_data.advance()
            
        retval = dict()
        
        if 'wavefield' in return_parameters:
            retval['wavefield'] = us
        if 'dWaveOp' in return_parameters:
            retval['dWaveOp'] = dWaveOp
        if 'simdata' in return_parameters:
            retval['simdata'] = simdata
            
        return retval

    @benchmark
    def solve(self, material, source_position, receiver_positions,
              wavelet_type="Ricker", source_fct=None, update_shot=False):
        
        shot = self.createShot(source_position, receiver_positions, wavelet_type, source_fct)
        base_model = self._solver.ModelParameters(self._mesh, {'C': material})
        
        self._solver.model_parameters = base_model
        ts = self._solver.ts()
        shot.reset_time_series(ts)
        
        shot.dt = self._solver.dt
        shot.trange = self._solver.trange
        
        #retval = self._objective.modeling_tools.forward_model(shot, base_model,
        #                                                      return_parameters=['dWaveOp', 'simdata', 'wavefield']) 
        
        retval = self._forward_model(shot, base_model,      
                                     return_parameters=['dWaveOp','simdata','wavefield'],
                                     update_shot=update_shot) 
        
        #generate_shot_data_time(shot, self._solver, base_model, verbose=False)	
        # #'simdata' ist gemessen, 'dWaveOp' das (2-te)Ableitungsfeld. 'wavefield' das volle Wellenfeld das alte shots[0].gather(as_array=True) quasi
        measurements = np.asarray(retval['simdata'])
        wavefields = retval['wavefield']
        
        return wavefields, measurements, shot, retval


    @benchmark
    def backpropAndGrad(self, shot, m, s_k, u_kdWaveOp):
        # print len(self._solver.ts())
        s_obs = shot.receivers.interpolate_data(self._solver.ts())
        s_residual = s_obs - s_k
        
        print "misfit: ", 0.5 * np.sum(s_residual**2) * self._solver.dt
        
        model_k = self._solver.ModelParameters(self._mesh, {'C': m})
        
        adjointfield = []
        g = self._objective.modeling_tools.migrate_shot(shot, model_k, s_residual,
                                                        dWaveOp=u_kdWaveOp,
                                                        adjointfield=adjointfield)
        
        return g.C, adjointfield

    @benchmark
    def caluclatealpha(self, shot, m, g, u_kdWaveOp):
        model_k = self._solver.ModelParameters(self._mesh, {'C': m})
        gradient = model_k.perturbation(data=g)
        linear_retval = self._objective.modeling_tools.linear_forward_model(shot, model_k, gradient,
                                                                            dWaveOp0=u_kdWaveOp,
                                                                            return_parameters=['simdata'])        
        delta_t = self._solver.dt
        norm_lin_sq = np.sum(linear_retval['simdata']**2) * delta_t # ???? * (delta_t*delta_t) * 7.5
        hsquared = np.prod(self._mesh.deltas)
        norm_adj_sq = np.sum(g**2) * hsquared
        alpha = norm_adj_sq / norm_lin_sq
        
        print "hsquared", hsquared, "dt", delta_t
        print "norm_adj", np.sqrt(norm_adj_sq)
        print "norm_lin", np.sqrt(norm_lin_sq)
        print "alpha", alpha
        
        return alpha


    def mesh(self): return self._mesh
    def domain(self): return self._domain

    @benchmark
    def plotSeismogram(self, measurements, figure_size=(100, 100), data_factor=2, color="black", draw_baselines=False, draw_receiver_numbers=False, linefactor=0.3):
        measurements = np.sqrt(np.abs(measurements)) * np.sign(measurements)
        
        fig = matplotlib.figure.Figure(figsize=(figure_size[0], figure_size[1]), dpi=1)
        canvas = FigureCanvas(fig)
        fig.patch.set_alpha(0.0)

        ax = fig.add_axes([0, 0, 1, 1])
        ax.set_axis_off()
        ax.patch.set_alpha(0.0)

        data = measurements.copy()
        data_range = np.min(data), np.max(data)
        R = data.shape[1]

        amplitude = max(map(abs, data_range))
        ym, yM = (0, 2 * (R - 1) * amplitude)
        dY = (yM - ym) / (R - 1)
        y_limits = ym, yM

        y_bnd_offsets = (-data_factor * amplitude, data_factor * amplitude)
        ax.set_ylim(y_limits[0] + y_bnd_offsets[0], y_limits[1] + y_bnd_offsets[1])

        ax.set_xlim([0, data.shape[0] - 1])

        m, M = data_range
        absM = max(abs(m), abs(M))

        data = data * data_factor * dY / (2 * absM)

        for i in range(R):
            tr = data[:, i].copy()
            offset_y = i * dY
            tr_offsetted = tr + offset_y
            # auxiliary lines
            if draw_baselines:
                ax.plot(np.ones(len(tr)) * offset_y, color="black", linewidth=0.05 * figure_size[1])
            # actual graphs
            ax.plot(tr_offsetted, color=color, linewidth=linefactor * figure_size[1])
            
            if draw_receiver_numbers and i+1 in (1,R,(R+2)/2) :
                ax.text(5,offset_y, "{:2d}".format(i+1), fontsize=1200,
                        horizontalalignment='left',
                        verticalalignment='center',
                        bbox=dict(facecolor='none', edgecolor='none', pad=2.0))

        fig.canvas.draw_idle()
        argb_data = bytes(fig.canvas.buffer_rgba())

#         qImage = QImage(argb_data, fig.canvas.renderer.width, fig.canvas.renderer.height, QImage.Format_ARGB32)
        qImage = QImage(argb_data, fig.canvas.renderer.width, fig.canvas.renderer.height, QImage.Format_RGBA8888)
        if DEBUG:
            qImage.save(ResPath.tmp + "latest_seismo.png")
        return qImage

    ##############################################
    # low level stuff    
    @benchmark
    def animate(self, wavefields, pause=1, frame_count=None, colormap=None):
        # calculate min and max values
        cMin, cMax = min([d.min() for d in wavefields]), max([d.max() for d in wavefields])

        if cMax - cMin == 0:
            mid = 0.0
        else:
            mid = abs(cMin) / (cMax - cMin)

#         gradient = ((0.0, 0.0, 0.0), (mid, 1.0, 1.0), (1.0, 0.0, 0.0))
#         cDict = dict.fromkeys(("red", "green", "blue"), gradient)
#         colormap = cm.colors.LinearSegmentedColormap("binaryNormed", cDict, 1024)
#         cm.register_cmap(cmap=colormap)

        if colormap is None:
            colormap = CenteredColorMap(mid, (0, 0, 1, 1), (0, 0, 1, 0), (1, 0, 0, 0), (1, 0, 0, 1), stretch=0.6)

        fig = plt.figure()
        fig.patch.set_alpha(0.0)
        plt.set_cmap(colormap)
        plt.clf()
        ax = fig.add_axes([0, 0, 1, 1])
        ax.set_axis_off()
        ax.patch.set_alpha(0.0)

        mesh = self.mesh()
        im = vis.plot(wavefields[0], mesh)
        im.set_clim((cMin, cMax))

        if frame_count is None:
            step = 1
        else:
            step = max(int(len(wavefields) / frame_count), 1)

        frames = []
        for i in range(0, len(wavefields), step):
            d = wavefields[i].copy()          

            m, M = np.min(d), np.max(d)
            m, M = cMin, cMax
            d[(d < 0.005 * M) * (d > 0)] = 0.0
            d[(d > 0.005 * m) * (d < 0)] = 0.0

            vis.plot(d, mesh, axis=ax, update=im)
            
            plt.draw()
            # argb_data = fig.canvas.tostring_argb()
            argb_data = bytes(fig.canvas.buffer_rgba())
            qImage = QImage(argb_data, fig.canvas.renderer.width, fig.canvas.renderer.height, QImage.Format_ARGB32)

            # qImage.save("/tmp/AAAA{:05d}.jpg".format(i))
            frames.append(qImage)

        plt.close(fig)
        return frames
    
    def animate_generator(self, wavefields, pause=1, frame_count=None, colormap=None):
        # calculate min and max values
        cMin, cMax = min([d.min() for d in wavefields]), max([d.max() for d in wavefields])

        if cMax - cMin == 0:
            mid = 0.0
        else:
            mid = abs(cMin) / (cMax - cMin)

#         gradient = ((0.0, 0.0, 0.0), (mid, 1.0, 1.0), (1.0, 0.0, 0.0))
#         cDict = dict.fromkeys(("red", "green", "blue"), gradient)
#         colormap = cm.colors.LinearSegmentedColormap("binaryNormed", cDict, 1024)
#         cm.register_cmap(cmap=colormap)

        if colormap is None:
            colormap = CenteredColorMap(mid, (0, 0, 1, 1), (0, 0, 1, 0), (1, 0, 0, 0), (1, 0, 0, 1), stretch=0.6)

        fig = plt.figure()
        fig.patch.set_alpha(0.0)
        plt.set_cmap(colormap)
        plt.clf()
        ax = fig.add_axes([0, 0, 1, 1])
        ax.set_axis_off()
        ax.patch.set_alpha(0.0)

        mesh = self.mesh()
        im = vis.plot(wavefields[0], mesh)
        im.set_clim((cMin, cMax))

        if frame_count is None:
            step = 1
        else:
            step = max(int(len(wavefields) / frame_count), 1)

        frames = []
        for i in range(0, len(wavefields), step):
            d = wavefields[i].copy()          

            m, M = np.min(d), np.max(d)
            m, M = cMin, cMax
            d[(d < 0.005 * M) * (d > 0)] = 0.0
            d[(d > 0.005 * m) * (d < 0)] = 0.0

            vis.plot(d, mesh, axis=ax, update=im)
            
            plt.draw()
            argb_data = bytes(fig.canvas.buffer_rgba())
            qImage = QImage(argb_data, fig.canvas.renderer.width, fig.canvas.renderer.height, QImage.Format_ARGB32)

            yield qImage

        plt.close(fig)


class FrameGenerator(object):
    def __init__(self, fwd, wavefields, frame_count=None, colormap=None):
        self._images = []
        self._wavefields = wavefields
        self._k = 0
        self._frame_count = frame_count
        self._animate_generator = fwd.animate_generator(self._wavefields, frame_count=self._frame_count,
                                                              colormap=colormap)
        self._finished = False
        next(iter(self))
    
    def __iter__(self):
        return self
    
    def __len__(self):
        return self._frame_count
    
    def finished_frame_count(self):
        return len(self._images)
    
    def rendering_finished(self):
        return self._finished
    
    def next(self):        
        if self._k >= self._frame_count:
            self._k = 0
            self._finished = True
            raise StopIteration
        
        if self._k >= len(self._images):
            try:
                frame = next(self._animate_generator)
                self._images += [frame]
            except StopIteration:
                self._frame_count = len(self._images)
                self._k = 0
                self._finished = True
                raise StopIteration
            
            self._k += 1
            
    def __getitem__(self, k):
        if k >= len(self._images):
            return self._images[-1]
        return self._images[k]
            



if __name__ == "__main__":
    def create_example_material(mesh, domain):
        def _gaussian_derivative_pulse(XX, YY, threshold, **kwargs):
            """ Derivative of a Gaussian at a specific sigma """
            T = -5 * YY * np.exp(-(YY ** 2 * 10))
            T[np.where(abs(T) < threshold)] = 0
            return T

        [XX, YY] = mesh.mesh_coords()
        material = 1 * np.ones(mesh.shape())

        material += _gaussian_derivative_pulse(XX, YY - 30, 1e-7)
        return 10 * material

    source_pos = (40.13251783893986, 6.477272727272727)
    receiver_pos = [(52.330275229357795, 4.534090909090909), (55.869520897043834, 4.534090909090909), (59.408766564729866, 4.534090909090909), (62.9480122324159, 4.534090909090909), (66.48725790010194, 4.534090909090909), (70.02650356778797, 4.534090909090909), (73.56574923547402, 4.534090909090909)]

    # N = 10
    # receiver_pos = [(52.330275229357795 + i*(73.56574923547402-52.330275229357795)/(N-1), 4.534090909090909) for i in range(N)]

    FWD = ForwardSolver(
        dimensions=((0.0, 100.0), (0.0, 70.0)), grid_resolution=(70, 100), trange=(0, 10),
        )

    mesh, domain = FWD.mesh(), FWD.domain()
    material = create_example_material(mesh, domain)
    wavefields, s_obs, shot, retval = FWD.solve(material, source_position=source_pos, receiver_positions=receiver_pos,
                                                update_shot=True)
    frames = FWD.animate(wavefields, frame_count=30)

    shot.receivers.data = s_obs
    print(shot)

    FWD.plotSeismogram(s_obs, (800, 300)).save("/tmp/seismogram_real.png")
    FWD.plotSeismogram(shot.receivers.interpolate_data(FWD._solver.ts()), (800, 300)).save("/tmp/seismogram_real_XXXX.png")


    for i, f in enumerate(frames):
        f.save("/tmp/REAL{:05d}.png".format(i))

#     IWS = InverseSolver(shot, 10*np.ones_like(material), FWD, source_pos, receiver_pos)
    # IWS = InverseSolver(shot, material.copy(), FWD, source_pos, receiver_pos)
    
    # inv_solv_thread = InverseSolverThread(IWS)
    # inv_solv_thread.start()
    
#     IWS.start()
#     import time
#     time.sleep(10)
#     IWS.stop()
    
    # IWS.join()


