from PIL import Image
from PyQt5.QtCore import QBuffer, QIODevice, QPoint, QRect
from PyQt5.QtGui import QImage, QPainter
from lib.constants import Orientation, COLOR_ROCK, COLOR_OIL, COLOR_BACKGROUND
from scipy.ndimage import filters
import cStringIO
import numpy as np
from util.decorators import benchmark
import matplotlib 
from PIL.ImageQt import ImageQt


def qImageToPillow(qtImage):
    """
    converts a QImage to a PIL-Image.
    @param qtImage: the QImage
    @return: a Pillow image
    """
    buf = QBuffer()
    buf.open(QIODevice.ReadWrite)
    qtImage.save(buf, "PNG")
    strIO = cStringIO.StringIO()
    strIO.write(buf.data())
    buf.close()
    strIO.seek(0)
    return Image.open(strIO)

def qImageToArray(img, colorTable, basis_value=0.0, weights=(1.0, 0.5), factor=30):
    """
    converts a QImage object to a numpy array.
    @param img: the QImage to convert
    @param colorTable: the color Table, it should be a tuple of 2 RGB-color-tuples: (mask1Color, mask2Color)
    @keyword basis_value: value to add to the whole array
    @keyword weights: a tuple of weights for the corresponding color
    @keyword factor: the factor to scale the end result with
    @return: a numpy-array representing the QImage, where all entries with color colorTable[i] have the value weights[i] (i=0,1), else value 0
    """
    pilImage = qImageToPillow(img)
    
    arr = np.asarray(pilImage, dtype=np.float)

    # filter out all hard material pixels
    hardMask = arr == colorTable[0]
    hardMask = np.all(hardMask, axis=2)
    # filter out all soft material pixels
    softMask = arr == colorTable[1]
    softMask = np.all(softMask, axis=2)
    # multiply masks with suitable values and add them
    arr = hardMask * (weights[0] * factor) + softMask * (weights[1] * factor)
    arr += basis_value * factor
    
    arr = filters.gaussian_filter(arr, sigma=0.8, order=0)
    arr = arr.T
    # debug
#     misc.imsave("../tmp/smooth.png", arr)
    return arr

@benchmark
def arrayToQImage(arr, colortable=None, basis_value=0.0, weights=(1.0, 0.5), factor=30):
    '''
    take care, some parameters currently aren't in use.
    '''
    # these three lines enable to see only the 10% of the highes values.
    # mean = np.percentile(arr, 90)
    # arr[arr > mean] = weights[0] * factor + basis_value * factor
    # arr[arr <= mean] = basis_value * factor
    
    bgColor = np.array(COLOR_BACKGROUND) / 255.0
    softColor = np.array(COLOR_OIL) / 255.0
    hardColor = np.array(COLOR_ROCK) / 255.0
    
    base = basis_value * factor
    high = weights[0] * factor + base
    mid = weights[1] * factor + base
    cmap_data = [(0, bgColor), (1.0, hardColor/5)] #, (0.5, softColor)] # HACKY...
    cmap = matplotlib.colors.LinearSegmentedColormap.from_list('arrToMaterial', cmap_data , N=256)
    # cmap.set_under(color=(0.0,0.0,1.0))
    sm = matplotlib.cm.ScalarMappable(cmap=cmap)    
    
    sm.set_clim(base, high)   
    res = sm.to_rgba(arr.T, bytes=True)
    pilImage = Image.fromarray(res)
    return ImageQt(pilImage)
    
    
    

def createFilledImage(srcImage, targetSize, tilesX=1, tilesY=1):
    """
    creates a image of specified size that is filled with a given image.
    @param srcImage: the QImage to repeat
    @param targetSize: the desired image (-output) size as QSize
    @param tilesX: into how many parts should the targetRect be divided horizontally,
                    any value <= 0 means that this adjusts the srcImage width by as little as possible to repeat it inside the targetRect
    @param tilesY: into how many parts should the targetRect be divided vertically,
                    any value <= 0 means that this adjusts the srcImage height by as little as possible to repeat it inside the targetRect
    @return: the filled QImage
    """
    if tilesX <= 0:
        tilesX = max(1, targetSize.width() // srcImage.width())
    if tilesY <= 0:
        tilesY = max(1, targetSize.height() // srcImage.height())

    tileWidth = targetSize.width() // tilesX
    tileHeight = targetSize.height() // tilesY
    if tileWidth * tilesX != targetSize.width() or tileHeight * tilesY != targetSize.height():
        print "Warning, targetSize {} is no multiple of tileSize {}".format((tileWidth, tileHeight), (targetSize.width(), targetSize.height()))
    image = srcImage.scaled(tileWidth, tileHeight)  # , transformMode=Qt.SmoothTransformation)

    outputImage = QImage(targetSize, srcImage.format())
    outputImage.fill(0)
    painter = QPainter(outputImage)

    for iy in range(tilesY):
        for ix in range(tilesX):
            painter.drawImage(QPoint(ix * tileWidth, iy * tileHeight), image)

    return outputImage

def mergeImages(images, orientation=Orientation.VERTICAL):
    """
    clips images together along the specified orientation.
    @param images: an iterable of QImages
    @keyword orientation: merge orientation, either Orientation.VERTICAL or Orientation.HORIZONTAL. Default is Orientation.VERTICAL.
    @return: the output QImage
    the width or height (depends on the orientation) of the output image corresponds to the width or height of the first image.
    """
    width = images[0].width()
    height = images[0].height()

    for image in images[1:]:
        if orientation == Orientation.HORIZONTAL:
            width += image.width()
        elif orientation == Orientation.VERTICAL:
            height += image.height()

    outputImage = QImage(width, height, images[0].format())
    outputImage.fill(0)
    painter = QPainter(outputImage)
    x = 0
    y = 0

    for image in images:
        if orientation == Orientation.HORIZONTAL:
            painter.drawImage(QRect(x, y, image.width(), height), image)
            x += image.width()
        elif orientation == Orientation.VERTICAL:
            painter.drawImage(QRect(x, y, width, image.height()), image)
            y += image.height()
            
    painter.end()

    return outputImage

def qImageToHash(qImage):
    pilImage = qImageToPillow(qImage)
    arr = np.asarray(pilImage, dtype=np.float)
    return hash(bytes(arr.data))
    
#     ptr = qImage.bits()
#     ptr.setsize(qImage.byteCount())
#     return hash(ptr)
