from multiprocessing import Lock

from lib.constants import Orientation
import lib.constants as constants
from lib.pysit_interface import calculateGridSize, ForwardSolver
from util import renderutil
from widgets.simulation.seismogramwidget import SeismogramWidget


class SimulationContext:
    def __init__(self):
        # set in materialeditortab.MaterialEditorWidget.resizeEvent
        self._gridWidth = 0
        self._gridWaterHeight = 0
        self._gridMaterialHeight = 0
        self.emitterPosition = None
        self.receiverPositions = None
        # set in materialeditortab.MaterialEditorWidget.openWaveDialog
        self.sourceFunction = None
        # set in simulationtab.AnimationWidget.resizeEvent
        self.seismogramWidth = None

        self._fwd = None
        # set in materialeditortab.MaterialEditorWidget.onStartSimulation
        self.material = None
        self.initialMaterial = None

        # created in materialeditortab.MaterialEditorWidget.onStartSimulation
        self.fwdResult = None

    @property
    def gridWidth(self):
        return self._gridWidth

    @property
    def waterHeight(self):
        return self._gridWaterHeight

    @property
    def materialHeight(self):
        return self._gridMaterialHeight

    @property
    def simulationHeight(self):
        return self._gridWaterHeight + self._gridMaterialHeight

    @property
    def seismogramSize(self):
        return self.seismogramWidth, SeismogramWidget.HEIGHT

    ##################
    # Forward Solver #
    ##################
    def _createFWDSolver(self):
        dim = ((0, self._gridWidth), (0, self.simulationHeight))
        res = (self._gridWidth, self.simulationHeight)
        return ForwardSolver(dimensions=dim, grid_resolution=res, trange=(0, constants.DEFAULT_TIME_RANGE))
    
    @property
    def FWDSolver(self):
        if not self._fwd:
            self._fwd = self._createFWDSolver()
        return self._fwd

    def updateGridSize(self, width, waterHeight, materialHeight):
        self._gridWidth, self._gridWaterHeight, self._gridMaterialHeight = calculateGridSize(width, waterHeight, materialHeight)
        # create new FWDSolver
        self._fwd = self._createFWDSolver()

class FWDResult:
    """
    Container for everything corresponding to the results of a forward simulation
    """
    def __init__(self, shot, measurements, realMaterialImage, signalPlacementImage, frameGenerator, seismogramImageBlack, seismogramImageGray):
        self._shot = shot
        self._measurements = measurements
        
        self._sceneImage = renderutil.mergeImages((signalPlacementImage, realMaterialImage), Orientation.VERTICAL)
        self._realMaterialImage = realMaterialImage
        self._signalPlacementImage = signalPlacementImage
        self._frameGenerator = frameGenerator
        self._seismogramImageBlack = seismogramImageBlack
        self._seismogramImageGray = seismogramImageGray
        
    @property
    def shot(self):
        return self._shot
    
    @property
    def measurements(self):
        return self._measurements
    
    @property
    def sceneImage(self):
        return self._sceneImage

    @property
    def realMaterialImage(self):
        return self._realMaterialImage
    
    @property
    def signalPlacementImage(self):
        return self._signalPlacementImage
    
    @property
    def frameGenerator(self):
        return self._frameGenerator
    
    @property
    def seismogramImageBlack(self):
        return self._seismogramImageBlack
    
    @property
    def seismogramImageGray(self):
        return self._seismogramImageGray

class InterateContainer:
    def __init__(self):
        self.lock = Lock()
        self.iterates = {}
        
    def __len__(self):
        with self.lock:
            return len(self.iterates)
        
    def __getitem__(self, k):
        with self.lock:
            if k in self.iterates:
                return self.iterates[k]
            else:
                return None

    def __setitem__(self, k, iterate):
        with self.lock:
            self.iterates[k] = iterate
            
    def __delitem__(self, k):
        with self.lock:
            del self.iterates[k]
            
    def clear(self):
        with self.lock:
            self.iterates.clear()
            
    
class InversionResult:
    """
    Container for everything corresponding to the results of a inversion 
    """
    def __init__(self):
        self.forward_solutions = InterateContainer()
        self.backward_solutions = InterateContainer()
        self.material_images = InterateContainer()
        self.material_iterates = InterateContainer()
        self.measurement_iterates = InterateContainer()
        self.seismogram_iterates = InterateContainer()
    
        self._all_containers = (self.forward_solutions, self.backward_solutions, self.material_iterates,
                                self.measurement_iterates, self.seismogram_iterates)
    
    def clear(self):
        for c in self._all_containers:
            c.clear()
            
    def isEmpty(self):
        return sum(len(c) for c in self._all_containers) == 0
