from PyQt5.QtCore import QCoreApplication, QPoint, QSize, Qt
from PyQt5.QtGui import QColor, QIcon, QImage, QPainter
from PyQt5.QtWidgets import QAction, QActionGroup, QGridLayout, QPushButton, QSizePolicy, QSpacerItem, QToolBar, QWidget
from lib.constants import Orientation, ResPath
from lib.i18n import lu
from lib.pysit_interface import FrameGenerator, WATER_MATERIAL, BACKGROUND_MATERIAL
from util import renderutil
from widgets.materialeditor.canvas import Canvas
from widgets.materialeditor.gallerywidget import GalleryWidget
from widgets.materialeditor.signalplacement import SignalPlacement
from util.simulationcontext import FWDResult
import numpy as np


class MaterialEditorWidget(QWidget):
    def __init__(self, handle, parent=None):
        QWidget.__init__(self, parent)
        self._handle = handle

        # app layout
        gridLayout = QGridLayout(self)
        gridLayout.setVerticalSpacing(0)  # we don't want vertical space between the widgets

        # emitter / receivers widget + water
        self._signalPlacement = SignalPlacement()
        self._signalPlacement.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)
        gridLayout.addWidget(self._signalPlacement, 0, 1, 1, 2)

        # canvas
        self._canvas = Canvas()
        self._canvas.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        gridLayout.addWidget(self._canvas, 1, 1, 1, 2)

        # Toolbar
        toolbar = QToolBar(orientation=Qt.Vertical)
        toolbar.setIconSize(QSize(32, 32))
        gridLayout.addWidget(toolbar, 1, 0)
        # actions for draw mode
        penGroup = QActionGroup(self)
        self._aHardMat = QAction(QIcon(ResPath.icons + "rock.png"), lu("actionSwitchToHardMaterial"), penGroup, triggered=self._canvas.activatePencil, checkable=True, checked=True)
        #QAction(QIcon(ResPath.icons + "oil.png"), lu("actionSwitchToSoftMaterial"), penGroup, triggered=lambda: self._canvas.activatePencil(hard=False), checkable=True)
        QAction(QIcon(ResPath.icons + "eraser.png"), lu("actionSwitchToEraseMode"), penGroup, shortcut="E", triggered=lambda: self._canvas.setMode(Canvas.MODE_ERASER), checkable=True)
        QAction(QIcon(ResPath.icons + "fill.png"), lu("actionSwitchToFillMode"), penGroup, shortcut="F", triggered=lambda: self._canvas.setMode(Canvas.MODE_FILL), checkable=True)
        toolbar.addActions(penGroup.actions())
        # clear action
        toolbar.addAction(QAction(QIcon(ResPath.icons + "clear.png"), lu("actionClearDrawingBoard"), self, shortcut="C", triggered=self._canvas.clear))

        # space below canvas
        gridLayout.addItem(QSpacerItem(0, 10), 2, 0, 1, 3)
        # example gallery
        self._gallery = GalleryWidget(title=lu("labelExamples") + ":", thumbCallback=self._canvas.setImage)
        self._gallery.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)
        self._gallery.setMinimumSize(100, 200)
        self._gallery.setAutoFillBackground(True)
        gridLayout.addWidget(self._gallery, 3, 1, 1, 2)

        # space below examples
        gridLayout.addItem(QSpacerItem(0, 10), 4, 0, 1, 2)
        # help  button
        bHelp = QPushButton("?", clicked=self.showHelp)
        bHelp.setMaximumWidth(toolbar.sizeHint().width())
#         bHelp.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        gridLayout.addWidget(bHelp, 5, 0)
        # start simulation button
        bStartSim = QPushButton(lu("buttonStartSimulation"), clicked=self.onStartSimulation)
        bStartSim.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        gridLayout.addWidget(bStartSim, 5, 1)
        # wave config button
#         self._bWaveConfig = QPushButton(icon=QIcon(ResPath.icons + "gear.png"), clicked=self.openWaveDialog, enabled=False)
#         self._bWaveConfig.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
#         gridLayout.addWidget(self._bWaveConfig, 5, 2)
        

    def resizeEvent(self, event):
        QWidget.resizeEvent(self, event)
        width = self._canvas.width()
        waterHeight = self._signalPlacement.waterHeight()
        materialHeight = self._canvas.height()

        ctx = self._handle.simulationContext
        ctx.updateGridSize(width, waterHeight, materialHeight)
        self._canvas.resetImage(ctx.gridWidth, ctx.materialHeight)

    def onStartSimulation(self):
        # stop inversion
        self._handle._inversionWidget.reset()
        
        # switch documentation page TODO maybe don't hard-code this
        self._handle._documentationTab.setCurrentIndex(0)
        self._handle._documentationTab._docWidgetTut.goTo(2)
        
        # show 'loading'-overlay
        self._handle.getOverlay().addText((self.width() / 2, self.height() / 2), lu("overlayLoading"))
        self._handle.getOverlay().show(animate=False)
        QCoreApplication.instance().processEvents()

        signalPlacementImage = QImage(self._signalPlacement.size(), QImage.Format_ARGB32)
        self._signalPlacement.render(QPainter(signalPlacementImage))
        canvasImage = QImage(self._canvas.size(), QImage.Format_ARGB32)
        self._canvas.render(QPainter(canvasImage))
        # save current drawn configuration
        self._canvas._image.save("../tmp/materialConfig.png")
        self._gallery.addLatestUserMaterial(self._canvas._image)

        # pysit
        ctx = self._handle.simulationContext

        material = WATER_MATERIAL * np.ones((ctx.gridWidth, ctx.simulationHeight), dtype=np.float)
        material[:, ctx.waterHeight:] = BACKGROUND_MATERIAL
        
        bottom_material = renderutil.qImageToArray(self._canvas.getRGBImage(), self._canvas.getColorTable(), basis_value=0.0)
        material[:, ctx.waterHeight:] += bottom_material
        
        material_flattened = material.reshape((ctx.gridWidth * ctx.simulationHeight, 1))

        # set context parameters
        ctx.material = material_flattened
        ctx.emitterPosition = self._signalPlacement.emitterPosition(ctx.gridWidth, ctx.waterHeight)
        ctx.receiverPositions = self._signalPlacement.receiversPositions(ctx.gridWidth, ctx.waterHeight)

        # solve forward problem
        wavefields, measurements, shot , __ = ctx.FWDSolver.solve(material_flattened,
                                                        source_position=(ctx.emitterPosition[0], ctx.emitterPosition[1]),
                                                        receiver_positions=ctx.receiverPositions, update_shot=True)
        # create seismogram image
        seismogramImageBlack = ctx.FWDSolver.plotSeismogram(measurements, figure_size=ctx.seismogramSize,
                                                       draw_receiver_numbers=True)
        seismogramImageGray = ctx.FWDSolver.plotSeismogram(measurements, figure_size=ctx.seismogramSize,
                                                       draw_receiver_numbers=True, color="888888")
        
        
        # create frame generator
        frameGenerator = FrameGenerator(ctx.FWDSolver, wavefields, frame_count=100)

        background_material = material.copy()
        background_material[:, ctx.waterHeight:] = renderutil.qImageToArray(self._canvas.getEmptyRGBImage(), self._canvas.getColorTable(), basis_value=0.15)

        # set remaining context parameters
        ctx.initialMaterial = background_material
        ctx.fwdResult = FWDResult(shot, measurements, canvasImage, signalPlacementImage, frameGenerator, seismogramImageBlack, seismogramImageGray)

        # switch to simulation tab
        self._handle.switchToSimulation()
        # hide overlay after calculation
        self._handle.getOverlay().hide(animate=False)

#     def onStartSimulation(self):
#         # show 'loading'-overlay
#         self._handle.getOverlay().addText((QPoint(self.width() / 2, self.height() / 2),), ("Loading",))
#         self._handle.getOverlay().show(animate=False)
#         QCoreApplication.instance().processEvents()
#
#         signalPlacementImage = QImage(self._signalPlacement.size(), QImage.Format_ARGB32)
#         self._signalPlacement.render(QPainter(signalPlacementImage))
#         canvasImage = QImage(self._canvas.size(), QImage.Format_ARGB32)
#         self._canvas.render(QPainter(canvasImage))
#         # save current drawn configuration
#         self._canvas._image.save("../tmp/materialConfig.png")
#
#         sceneImage = renderutil.mergeImages((signalPlacementImage, canvasImage), Orientation.VERTICAL)
#
#         # pysit
#         # material = float(self._simulationAreaTotalHeight)/5*np.ones((self._simulationAreaWidth, self._simulationAreaTotalHeight), dtype=np.float)
#         material = 10 * np.ones((self._simulationAreaWidth, self._simulationAreaTotalHeight), dtype=np.float)
#
# #         self._canvas.getRGBImage().save(ResPath.tmp + "materialbefore.png")
#
#         bottom_material = renderutil.qImageToArray(self._canvas.getRGBImage(), self._canvas.getColorTable(), basis_value=0.15)
#         material[:, self._simulationAreaWaterHeight:] += bottom_material
#         # test backwards
#
#         # TODO remove tmp
# #         reconstruction = renderutil.arrayToQImage(bottom_material, colortable=self._canvas.getColorTable(), basis_value=0.15, weights=(1.0, 0.5), factor=30)
# #         reconstruction.save(ResPath.tmp + "reconstruction.png")
# #         np.set_printoptions(precision=1, threshold=20000, edgeitems=1000000, linewidth=14000)
# #         with open(ResPath.tmp + "materialSimulation.txt", "w") as f:
# #             f.write(str(bottom_material))
#
#         material_flattened = material.reshape((self._simulationAreaWidth * self._simulationAreaTotalHeight, 1))
#
#         emitterPosition = self._signalPlacement.emitterPosition(self._simulationAreaWidth, self._simulationAreaTotalHeight)
#         receiversPositions = self._signalPlacement.receiversPositions(self._simulationAreaWidth, self._simulationAreaTotalHeight)
#
#         FWD = ForwardSolver(
#             dimensions=((0, self._simulationAreaWidth), (0, self._simulationAreaTotalHeight)),
#             grid_resolution=(self._simulationAreaWidth, self._simulationAreaTotalHeight),
#             trange=(0, 10))
# #         wavefields, measurements = FWD.solve(material_flattened,
# #                                              source_position=(emitterPosition[0], emitterPosition[1]),
# #                                              receiver_positions=receiversPositions,
# #                                              wavelet_type="custom", source_fct=source_fct)
#         wavefields, measurements, shot , __ = FWD.solve(material_flattened,
#                                                         source_position=(emitterPosition[0], emitterPosition[1]),
#                                                         receiver_positions=receiversPositions)
#         seismogram_image = FWD.plotSeismogram(measurements, figure_size=self._handle.seismogramSize())
#
#         background_material = material.copy()
#         background_material[:, self._simulationAreaWaterHeight:] = \
#             renderutil.qImageToArray(self._canvas.getEmptyRGBImage(),
#                                      self._canvas.getColorTable(), basis_value=0.15)
#
#         frame_generator = FrameGenerator(FWD, wavefields, frame_count=100)
#         self._handle.switchToSimulation(FWD, signalPlacementImage, material_flattened, background_material, shot, measurements,
#                                         emitterPosition, receiversPositions,
#                                         sceneImage, frame_generator, seismogram_image)
#
#         # hide overlay after calculation
#         self._handle.getOverlay().hide(animate=False)

    def openWaveDialog(self):
        from widgets.signalconfig.waveconfiguration import WaveDialog
        d = WaveDialog(self)
        if d.exec_():
            # that is the resulting function:
            # TODO use this function
            self._handle.simulationContext.sourceFunction = d.getResult()

    def showHelp(self):
        self._handle.getOverlay().addText((self.width() / 2, self.height() / 2), "TODO")
        self._handle.getOverlay().show()

    def reset(self):
        self._canvas.clear()
        # set hard material pencil as current tool
        self._aHardMat.setChecked(True)
        self._aHardMat.trigger()
        # disable wave config button
#         self._bWaveConfig.setEnabled(False)
        # maybe reset signalplacement
