from PyQt5.QtCore import QRect, Qt, QSize
from PyQt5.QtGui import QIcon, QPainter
from PyQt5.QtWidgets import QAction, QSizePolicy, QSlider, QSpacerItem, QToolBar, QWidget, QGridLayout, QPushButton
from lib.constants import ResPath
from lib.i18n import lu
from lib.inversion_solver import InverseSolverThread
from lib.pysit_interface import FrameGenerator, perturb_material
from util import renderutil
from widgets.materialeditor.signalplacement import SignalPlacement
from widgets.simulation.jumpslider import JumpSlider
from widgets.simulation.seismogramwidget import SeismogramWidget
import time


class InversionWidget(QWidget):
    def __init__(self, handle, parent=None):
        QWidget.__init__(self, parent)
        self._handle = handle
        self._lblcurrentiterationText = lu("labelCurrentIteration")
        self._iteration = 0
        # TODO reimplement inverted?
#         self._inverted = False
        self._material = None

        self._fwd_counter = 0
        self._bwd_counter = 0
        self._frame_generator = None

        self._inv_thread = None

        self._material_iteration = None
        self._seismogram_iteration = None

        #self._current_seismogram = None
        #self._optimal_seismogram = None
        # app layout
        gridLayout = QGridLayout(self)
        gridLayout.setVerticalSpacing(0)  # we don't want vertical space between the widgets

        # space for image
        gridLayout.addItem(QSpacerItem(0, 100, vPolicy=QSizePolicy.Expanding), 0, 0, 1, 2)

        # toolbar (play, pause)
        toolbar = QToolBar(orientation=Qt.Vertical)
        toolbar.setLayoutDirection(Qt.RightToLeft)
        toolbar.setIconSize(QSize(32, 32))
#         # stop action
#         toolbar.addAction(QAction(QIcon(ResPath.icons + "stop.png"), lu("actionStop"), self, triggered=self.stop))
        # play/pause action
        icon = QIcon(ResPath.icons + "play.png")
        icon.addFile(ResPath.icons + "pause.png", state=QIcon.On)
        trigger = lambda: None # TODO
        self._aPlayPause = QAction(icon, lu("actionPlay"), self, checkable=True, checked=True, shortcut=" ", triggered=trigger)
        toolbar.addAction(self._aPlayPause)

        self._buttonRealMaterial = QPushButton("\n".join(lu("buttonShowRealMaterial")), pressed=self.update, released=self.update)
        self._buttonRealMaterial.setMaximumWidth(toolbar.sizeHint().width())
        self._buttonRealMaterial.setFocusPolicy(Qt.NoFocus)
        gridLayout.addWidget(self._buttonRealMaterial, 0, 0)

        gridLayout.addWidget(toolbar, 1, 0, 2, 1, alignment=Qt.AlignBottom)

        # slider
        self.slider = JumpSlider(Qt.Horizontal, tickPosition=QSlider.TicksBelow, valueChanged=self.update)
        self.slider.setTickInterval(1)
        self.slider.setPageStep(1)
        # TODO: do this prettier
        nothing = lambda e: None
        self.slider.mouseMoveEvent = nothing
        self.slider.mousePressEvent = nothing
        self.slider.wheelEvent = nothing
        # seismogram widget
        self._seismogram = SeismogramWidget(self.slider)
        gridLayout.addWidget(self._seismogram, 1, 1)
        # have to add slider last
        gridLayout.addWidget(self.slider, 2, 1)
        # space below seismogram
        gridLayout.addItem(QSpacerItem(0, 8), 3, 0, 1, 2)

        self._buttonNextShipPosition = QPushButton(lu("buttonNextShipPosition"), clicked=self.selectNextShipPosition)
        gridLayout.addWidget(self._buttonNextShipPosition, 4, 1)

        self._state = "labelInitializing"
        
        self._timerId = None

#     def setImages(self, images):
#         self._images = images
#         self.slider.setMaximum(len(self._images) - 1)
#         if self._inverted:
#             self.slider.setValue(self.slider.maximum())
#         else:
#             self.slider.setValue(0)

#     def timerEvent(self, event):
#         finished = False
#         if self._inverted:
#             nextValue = (self.slider.value() - 1) % len(self._images)
#             if nextValue > self.slider.value() and not self._aLoop.isChecked():
#                 finished = True
#         else:
#             nextValue = (self.slider.value() + 1) % len(self._images)
#             if nextValue < self.slider.value() and not self._aLoop.isChecked():
#                 finished = True
#         self.slider.setValue(nextValue)
#         self.update()
#         if finished:
#             self.finished()

    def timerEvent(self, event):
        start = time.clock()

        if (self._INV is not None and self._INV.isEmpty() and 
                self._frame_generator is None and 
                self._inv_thread is not None and self._inv_thread.finished()):
            if self._timerId:
                self.killTimer(self._timerId)
                self._timerId = None
            
            self._seismogram.set_current_seismogram(None)
            self.inversionFinished()
            print "Rendering finished"

        # TODO:
        # Enable reset of material parameter
        # self.material_images = []
        # self.measurement_iterates = []
        # if self._bwd_counter <= new_material_index

        if self._INV.material_images[self._material_iteration] is not None:
            self._material = self._INV.material_images[self._material_iteration]
            del self._INV.material_images[self._material_iteration]
            print("new material!")
            
        if self._INV.material_iterates[self._material_iteration] is not None:
            self._material_array = self._INV.material_iterates[self._material_iteration]
            del self._INV.material_iterates[self._material_iteration]

        if self._frame_generator:
            self.slider.setValue(self.slider.value() + 1)

        if self._fwd_counter > self._bwd_counter:
            if self._frame_generator is None and self._INV.backward_solutions[self._bwd_counter] is not None:
                bwd_solution = self._INV.backward_solutions[self._bwd_counter][::-1]
                del self._INV.backward_solutions[self._bwd_counter]
                
                self._frame_generator = FrameGenerator(self._FWD, bwd_solution, frame_count=100)
                self.slider.setMaximum(len(self._frame_generator) - 1)
                self.slider.setValue(0)
                print "Initialized Backward Frame Generator"
                self._state = "labelBackwardSimulation"

            if self._frame_generator:
                if self.slider.value() == len(self._frame_generator) - 1:
                    self._bwd_counter += 1
                    self._frame_generator = None
                    self._material_iteration += 1
                    print "Deleted Backward Frame Generator"
        else:
            if self._frame_generator is None and self._INV.forward_solutions[self._fwd_counter] is not None:
                fwd_solution = self._INV.forward_solutions[self._fwd_counter]
                del self._INV.forward_solutions[self._fwd_counter]
                
                # new forwardsoltion and new seismogram!
                self._frame_generator = FrameGenerator(self._FWD, fwd_solution, frame_count=100)

                # if len(self._INV.measurement_iterates) > self._seismogram_iteration:
                #    self._seismogram_iteration += 1
                
                if self._INV.measurement_iterates[self._fwd_counter] is not None:
                    seismogram = self._INV.measurement_iterates[self._fwd_counter]
                    del self._INV.measurement_iterates[self._fwd_counter]
                    current_seismogram = self._FWD.plotSeismogram(seismogram,figure_size=self._seismogram_size, color="green")
                    self._seismogram.set_current_seismogram(current_seismogram)
                print("new seismogram!")
                #if DEBUG:
                #    self._current_seismogram.save(ResPath.tmp + "seismo_inv_{:03d}.png".format(self._seismogram_iteration))



                self.slider.setMaximum(len(self._frame_generator) - 1)
                self.slider.setValue(0)
                print "Initialized Forward Frame Generator"
                self._iteration += 1
                self._state = "labelForwardSimulation"
                self._seismogram.set_isForward(True)

            if self._frame_generator:
                if self.slider.value() == len(self._frame_generator) - 1:
                    self._seismogram.set_isForward(False)
                    self._fwd_counter += 1
                    self._frame_generator = None
                    print "Deleted Forward Frame Generator"

        if self._frame_generator:
            for _ in self._frame_generator:
                if time.clock() - start > 25 / 1000.0:
                    break

    def inversionFinished(self):
        ov = self._handle.getOverlay()
        ov.addText((100, self.height() / 2), lu("overlayMotivateSettingChange"))
        ov.show(opacity=50)
        
#     def toggle(self):
#         if not self._aPlayPause.isChecked():
#             self.start()
#         else:
#             self.pause()

#     def start(self):
#         if not self._timerId:
#             self._timerId = self.startTimer(40)
#             self._aPlayPause.setChecked(True)
#             self._aPlayPause.setText(lu("actionPause"))

#     def pause(self):
#         if self._timerId:
#             self.killTimer(self._timerId)
#             self._timerId = None
#             self._aPlayPause.setChecked(False)
#             self._aPlayPause.setText(lu("actionPlay"))

#     def stop(self):
#         self.pause()
#         if self._inverted:
#             self.slider.setValue(self.slider.maximum())
#         else:
#             self.slider.setValue(0)

#     def finished(self):
#         self.pause()

    def paintEvent(self, event):
        QWidget.paintEvent(self, event)
        rect = self.layout().cellRect(0, 1) # self.layout().itemAt(0).geometry()
        painter = QPainter(self)

        
        rect = rect.adjusted(0, 0, 0, -4)

        # draw water, ship and receivers
        if self._material and self._signalPlacementImage:
            signalsRect = QRect(rect.left(), rect.top(), rect.width(), self._signalPlacementImage.height())
            painter.drawImage(signalsRect, self._signalPlacementImage)

            drawn_material = (self._material_real if self._buttonRealMaterial.isDown() else self._material)
            painter.drawImage(rect.adjusted(0, self._signalPlacementImage.height(), 0, 0), drawn_material)

        
        if self._frame_generator:
            k = self.slider.value()
            while self._frame_generator.finished_frame_count() < k:
                next(self._frame_generator)

            # painter.setCompositionMode(QPainter.CompositionMode_Multiply)
            painter.drawImage(rect.adjusted(0, SignalPlacement.SHIP_HEIGHT, 0, 0), self._frame_generator[k])
        
        painter.setPen(Qt.white)
        text = self._lblcurrentiterationText + ": {}".format(self._iteration)
        painter.drawText(rect.adjusted(5, 0, 0, -5), Qt.AlignLeft | Qt.AlignBottom, text)
        
        state = lu("labelCurrentlyShowing") + ": " + lu(self._state)
        painter.drawText(rect.adjusted(0, 0, -5, -5), Qt.AlignRight | Qt.AlignBottom, state)
        
    def loadFromCurrentContext(self, overwrite_initial_material=None):
        ctx = self._handle.simulationContext

        width, wheight, totalheight = ctx.gridWidth, ctx.waterHeight, ctx.simulationHeight

        # modify real material to have somthing to calculate
        if overwrite_initial_material is None:
            initial_material = perturb_material(ctx.material.reshape(width, totalheight), ctx.waterHeight).reshape(*ctx.material.shape)
        else:
            initial_material = overwrite_initial_material

        # TODO 10 magic number from onStartSimulation
        self._material = renderutil.arrayToQImage(initial_material.reshape(width, totalheight)[:, wheight:] - 10)
        self._material_iteration = -1
        self._seismogram_iteration = 0

        self._FWD = ctx.FWDSolver
        self._seismogram_size = ctx.seismogramSize
        #self._material_real = renderutil.arrayToQImage(ctx.material.reshape(width, totalheight)[:, wheight:] - 10)
        self._material_real = ctx.fwdResult.realMaterialImage
        self._signalPlacementImage = ctx.fwdResult.signalPlacementImage
        self._seismogram.set_original_seismogram(ctx.fwdResult._seismogramImageBlack)

        #if DEBUG:
        #    self._optimal_seismogram.save(ResPath.tmp + "seismo_initial.png")

        # smoothed_material = real_material.copy()
        # smoothed_material = scipy.ndimage.filters.gaussian_filter(real_material, 20)
        # #self._inv_thread = InverseSolverThread(ctx.fwdResult.shot, initial_material, ctx.FWDSolver, ctx.emitterPosition, ctx.receiverPositions)

        self._inv_thread = InverseSolverThread(ctx, initial_material)
        self._INV = self._inv_thread.inv_solve.inv_result
        self._inv_thread.start()
        self._timerId = self.startTimer(30)

#     def initialize(self, sceneImage, signalPlacementImage, FWD, shot, source_pos, receiver_pos, real_material, initial_material, measurements):
#         #TODO better access to constants
#         self._material = renderutil.arrayToQImage(initial_material.reshape(121,57)[:,13:]-10)
#         self._material_iteration = 0
#         self._FWD = FWD
#         self._material_real = real_material
#         self._measurements = measurements
#         self._signalPlacementImage = signalPlacementImage
#         # smoothed_material = real_material.copy()
#         # smoothed_material = scipy.ndimage.filters.gaussian_filter(real_material, 20)
#
#         self._inv_thread = InverseSolverThread(shot, initial_material, FWD, source_pos, receiver_pos)
#         self._INV = self._inv_thread.inv_solve
#         self._inv_thread.start()
#         self.startTimer(50)

    def selectNextShipPosition(self):
        self._handle.keptInitialMaterial = self._material_array.copy()
        self._handle.switchToMaterialEditor()

    def reset(self):
#         self.stop()

        if self._timerId:
            self.killTimer(self._timerId)
            self._timerId = None

        if self._inv_thread:
            self._inv_thread.stop()
            
            print "stopping inversion thread..."
            
            # wait for the thread to terminate (UGLY!)
            while not self._inv_thread.finished():
                time.sleep(0.05)
                
            print "inversion thread finished"
            
        # TODO kill processes
        self._state = "labelInitializing"

        # reset counters and data
        self._iteration = 0
        self._material = None

        self._fwd_counter = 0
        self._bwd_counter = 0
        self._frame_generator = None

        self._inv_thread = None

        self._material_iteration = None
        self._seismogram_iteration = None

        self._current_seismogram = None
        self._optimal_seismogram = None

        # TODO is this allowed? Want to let GC take care of old inversion_data
        self._inv_thread = None
        self._INV = None
        self._frame_generator = None


    def cleanup(self):
        if self._inv_thread:
            self._inv_thread.stop()
