from PyQt5.QtCore import QSize, QTimer, Qt
from PyQt5.QtGui import QIcon, QPainter
from PyQt5.QtWidgets import QAction, QGridLayout, QPushButton, QSizePolicy, QSlider, QSpacerItem, QToolBar, QWidget
from jumpslider import JumpSlider
from lib.constants import ResPath
from lib.i18n import lu
from seismogramwidget import SeismogramWidget
from widgets.materialeditor.signalplacement import SignalPlacement
import time


class AnimationWidget(QWidget):
    def __init__(self, handle, parent=None):
        QWidget.__init__(self, parent)
        self._handle = handle

        self._timerId = None
        self.renderTimer = QTimer(timeout=self.renderTick)

        self._bg = None
        self._imageOffset = SignalPlacement.SHIP_HEIGHT

        # app layout
        gridLayout = QGridLayout(self)
        gridLayout.setVerticalSpacing(0)    # we don't want vertical space between the widgets

        # space for image
        gridLayout.addItem(QSpacerItem(0, 100, vPolicy=QSizePolicy.Expanding), 0, 0, 1, 2)

        # toolbar
        toolbar = QToolBar(orientation=Qt.Vertical)
        toolbar.setIconSize(QSize(32, 32))

        # loop action
        self._aLoop = QAction(QIcon(ResPath.icons + "loop.png"), lu("actionLoop"), self, checkable=True, checked=True)
        toolbar.addAction(self._aLoop)
        # stop action
        toolbar.addAction(QAction(QIcon(ResPath.icons + "stop.png"), lu("actionStop"), self, triggered=self.stop))
        # play/pause action
        icon = QIcon(ResPath.icons + "play.png")
        icon.addFile(ResPath.icons + "pause.png", state=QIcon.On)
        trigger = lambda: self.start() if self._aPlayPause.isChecked() else self.pause() # cant use self.toggle because action state is changed before call
        self._aPlayPause = QAction(icon, lu("actionPlay"), self, checkable=True, shortcut=" ", triggered=trigger)
        toolbar.addAction(self._aPlayPause)
        gridLayout.addWidget(toolbar, 1, 0, 2, 1, alignment=Qt.AlignBottom)

        # slider
        self.slider = JumpSlider(Qt.Horizontal, tickPosition=QSlider.TicksBelow, valueChanged=self.update)
        self.slider.setTickInterval(1)
        self.slider.setPageStep(1)
        self.slider.setInvertedControls(True)
        gridLayout.addWidget(self.slider, 2, 1)

        # seismogram widget
        self._seismogram = SeismogramWidget(self.slider, isForward = True)
        gridLayout.addWidget(self._seismogram, 1, 1)

        # space below seismogram
        gridLayout.addItem(QSpacerItem(0, 8), 3, 1)
        # start inversion button
        bStartInversion = QPushButton(lu("buttonStartInversion"), clicked=self.gotoInversion)
        gridLayout.addWidget(bStartInversion, 4, 1)
        
        self._runCounter = 3


    def resizeEvent(self, event):
        QWidget.resizeEvent(self, event)
        # update seismogramSize
        self._handle.simulationContext.seismogramWidth = self.width()

    def timerEvent(self, event):
        nextValue = (self.slider.value() + 1) % (len(self.frame_generator))
        if nextValue < self.slider.value():
            self._runCounter -= 1
            if not self._aLoop.isChecked() or self._runCounter <= 0:
                self.pause()
        if self._runCounter <= 0:
            self.slider.setValue((self.slider.maximum()-self.slider.minimum())//3)
        else:
            self.slider.setValue(nextValue)
        self.update()

    def renderTick(self):
        start = time.clock()
        for _ in self.frame_generator:
            if time.clock() - start > 25 / 1000.0:
                break

        if self.frame_generator.rendering_finished():
            self.renderTimer.stop()

    def toggle(self):
        if not self._aPlayPause.isChecked():
            self.start()
        else:
            self.pause()

    def start(self):
        if not self._timerId:
            self._runCounter = 3
            self._timerId = self.startTimer(50)
            self._aPlayPause.setChecked(True)
            self._aPlayPause.setText(lu("actionPause"))

    def pause(self):
        if self._timerId:
            self.killTimer(self._timerId)
            self._timerId = None
            self._aPlayPause.setChecked(False)
            self._aPlayPause.setText(lu("actionPlay"))

    def stop(self):
        self._runCounter = 3
        self.pause()
        self.slider.setValue(0)

    def _setBackground(self, bg):
        self._bg = bg
        cellHeight = self.layout().cellRect(0, 1).height() - 4
        self._imageOffset = int(round(cellHeight * SignalPlacement.SHIP_HEIGHT / float(self._bg.height())))
        self.update()

    def _setFrameGenerator(self, frame_generator):
        self.frame_generator = frame_generator
        self.slider.setMaximum(len(self.frame_generator) - 1)
        self.renderTimer.start(50)

    def loadFromCurrentContext(self):
        fwdResult = self._handle.simulationContext.fwdResult
        # set the background (scene) Image of this animation widget
        self._setBackground(fwdResult.sceneImage)
        # set the animation frames
        self._setFrameGenerator(fwdResult.frameGenerator)
        # set the seismogram
        self._seismogram.set_original_seismogram(fwdResult.seismogramImageGray)
        self._seismogram.set_current_seismogram(fwdResult.seismogramImageBlack)
    def wheelEvent(self, event):
        self.slider.wheelEvent(event)

    def paintEvent(self, event):
        QWidget.paintEvent(self, event)

        rect = self.layout().cellRect(0, 1).adjusted(0, 0, 0, -4)
        painter = QPainter(self)

        if self._bg:
            painter.drawImage(rect, self._bg)
        if self.frame_generator:
            k = self.slider.value()
            while self.frame_generator.finished_frame_count() < k:
                next(self.frame_generator)

            painter.drawImage(rect.adjusted(0, self._imageOffset, 0, 0), self.frame_generator[k])

        # painter.drawRect(rect.adjusted(-1, -1, 0, 0))

    def reset(self):
        self.stop()
        self._aLoop.setChecked(True)

    def gotoInversion(self):
        self.pause()
        
        # ensure that rendering is finished
        if self.frame_generator is not None:
            while not self.frame_generator.rendering_finished():
                try: 
                    next(self.frame_generator)
                except StopIteration:
                    break
        
        self._handle.switchToInversion()

#     def gotoInversion(self):
#         self._handle.switchToInversion(self._bg)
